from __future__ import annotations

import json
from pathlib import Path
from subprocess import PIPE, Popen
from typing import Any, AnyStr, Dict, List, Sequence

import openvds
from humanfriendly import format_size

from ovds_utils.logging import get_logger

logger = get_logger(__name__)


METADATATYPE_TO_OVDS_GET_FUNCTION = {
    "MetadataType.IntVector2": openvds.core.VolumeDataLayout.getMetadataIntVector2,
    "MetadataType.Float": openvds.core.VolumeDataLayout.getMetadataFloat,
    "MetadataType.DoubleVector2": openvds.core.VolumeDataLayout.getMetadataDoubleVector2,
    "MetadataType.String": openvds.core.VolumeDataLayout.getMetadataString,
    "MetadataType.Int": openvds.core.VolumeDataLayout.getMetadataInt,
    "MetadataType.Double": openvds.core.VolumeDataLayout.getMetadataDouble,
    "MetadataType.BLOB": openvds.core.VolumeDataLayout.getMetadataBLOB,
}


METADATATYPE_TO_OVDS_SET_FUNCTION = {
    "MetadataType.IntVector2": openvds.core.MetadataContainer.setMetadataIntVector2,
    "MetadataType.Float": openvds.core.MetadataContainer.setMetadataFloat,
    "MetadataType.DoubleVector2": openvds.core.MetadataContainer.setMetadataDoubleVector2,
    "MetadataType.String": openvds.core.MetadataContainer.setMetadataString,
    "MetadataType.Int": openvds.core.MetadataContainer.setMetadataInt,
    "MetadataType.Double": openvds.core.MetadataContainer.setMetadataDouble,
    "MetadataType.BLOB": openvds.core.MetadataContainer.setMetadataBLOB,
}


def get_bricksize_values(brick_size: openvds.core.VolumeDataLayoutDescriptor.BrickSize, shape: Sequence[int]):
    return [2 ** brick_size.value if s >= 2 ** brick_size.value else s for s in shape]


def get_element_size(format: openvds.core.VolumeDataFormat, components: openvds.core.VolumeDataComponents):
    if format == openvds.VolumeDataChannelDescriptor.Format.Format_1Bit:
        return 1
    elif format == openvds.VolumeDataChannelDescriptor.Format.Format_U8:
        return 1 * components.value
    elif format == openvds.VolumeDataChannelDescriptor.Format.Format_U16:
        return 2 * components.value
    elif format in (
        openvds.VolumeDataChannelDescriptor.Format.Format_R32,
        openvds.VolumeDataChannelDescriptor.Format.Format_U32,
    ):
        return 4 * components.value
    elif format in (
        openvds.VolumeDataChannelDescriptor.Format.Format_U64,
        openvds.VolumeDataChannelDescriptor.Format.Format_R64,
    ):
        return 8 * components.value
    else:
        raise Exception("Illegal format: ", format)


def check_block_size(
    brickSize: openvds.core.VolumeDataLayoutDescriptor.BrickSize,
    channels: int,
    shape: Sequence[int],
    format: openvds.core.VolumeDataFormat,
    components: openvds.core.VolumeDataComponents
):
    element_size = get_element_size(format, components)
    brick_size_values = get_bricksize_values(brickSize, shape)

    datablock_size = brick_size_values[0]
    for i in range(1, len(shape)):
        datablock_size *= brick_size_values[i]
    datablock_size *= channels * element_size
    logger.info(f"Datablock size: {format_size(datablock_size)}")
    logger.info(f"Estimated process size: {format_size(datablock_size*10)}")
    if datablock_size > 2147483647:
        raise Exception(
            f"Datablock is too big ({brick_size_values[0]} x {brick_size_values[1]}\
                 x {brick_size_values[2]} x {channels} x {element_size} bytes)"
        )


def copy_ovds_metadata(metadata_details: Dict[AnyStr, Any], result_metadata_container: openvds.core.MetadataContainer):
    for k, v in metadata_details.items():
        set_method = METADATATYPE_TO_OVDS_SET_FUNCTION[v.type]
        set_method(result_metadata_container, v.category, k, v.value)


def get_ovds_bin(name: str) -> str:
    """retrives absolute path to binary on openvds package"""

    import openvds as vds_package

    root_init_vds = Path(vds_package.__file__)
    assert root_init_vds.exists()
    assert root_init_vds.is_file()

    # build binary dir e.g. venv/bin
    bin_directory = root_init_vds.parent / ".." / ".." / ".." / ".." / "bin"
    assert bin_directory.exists()
    assert bin_directory.is_dir()

    # look for VDSInfo inside bin dir
    vdsinfo_bin = bin_directory / name
    assert vdsinfo_bin.exists()
    assert vdsinfo_bin.is_file()

    return str(vdsinfo_bin.resolve().absolute())


def get_vdsinfo_bin() -> str:
    return get_ovds_bin("VDSInfo")


def get_vds_info(path: AnyStr, connection_string: AnyStr):
    vds_bin = get_vdsinfo_bin()

    if connection_string:
        cmd = f'{vds_bin} {path} --connection "{connection_string}"'
    else:
        cmd = f"{vds_bin} {path} {connection_string}"
    process = Popen(cmd, stdout=PIPE, stderr=PIPE, shell=True)
    out, err = process.communicate()
    try:
        return json.loads(out)
    except Exception as e:
        logger.error(out)
        logger.error(err)
        raise Exception(f"Could not parse VDSInfo {e}")


def read_metadata_container(
    container: openvds.core.MetadataContainer,
) -> List[Dict[str, str]]:
    """serializes openvds.MetadataContainer to JSON entries"""
    ret = []
    for key in container.getMetadataKeys():
        ovds_getter = METADATATYPE_TO_OVDS_GET_FUNCTION[str(key.type)]
        value = ovds_getter(container, key.category, key.name)
        if not str(key.type).startswith("MetadataType."):
            raise RuntimeError(
                f"{key} has {key.type} malformed, expected 'MetadataType.' prefix !"
            )
        # NOTE: at this time
        # key.type should be 'MetadataType.X', we want to extract X,
        # so that later we can create correct oVDS type
        item_type = str(key.type).split(".")[1]
        ret.append(
            dict(
                category=key.category,
                name=key.name,
                value=value,
                type=item_type
            )
        )
    return ret


def create_metadata_container(
    entries: List[Dict[str, str]] = None
) -> openvds.core.MetadataContainer:
    """de-serializes JSON entries to openvds.MetadataContainer"""
    metadata_container = openvds.core.MetadataContainer()
    entries = sorted(entries, key=lambda x: x["category"])
    for entry in entries:
        ovds_setter = METADATATYPE_TO_OVDS_SET_FUNCTION[f"MetadataType.{entry['type']}"]
        ovds_setter(
            metadata_container, entry["category"], entry["name"], entry["value"]
        )

    return metadata_container



from __future__ import annotations

from copy import deepcopy
from typing import List, Sequence, Tuple, Union

import numpy as np
import openvds

from ovds_utils.exceptions import VDSException
from ovds_utils.logging import get_logger
from ovds_utils.metadata import MetadataContainer
from ovds_utils.ovds import (LOD, AccessModes, BrickSizes, Components, Dimensions, Formats, InitValue, Options,
                             create_vds)
from ovds_utils.ovds.utils import get_vds_info
from ovds_utils.ovds.writing import FORMAT2NPTYPE

logger = get_logger(__name__)


class VDSChunk:
    def __init__(
        self,
        number: int,
        accessor: openvds.core.VolumeDataPageAccessor,
        format: Formats
    ) -> None:
        super().__init__()
        self.is_released = False
        self.number = number
        self.accesor = accessor
        self._page = None
        self.format = format

    def __repr__(self) -> str:
        return f"<VDSChunk(number={self.number})>"

    def __getitem__(self, key: Sequence[Union[int, slice]]) -> np.array:
        dtype = FORMAT2NPTYPE[self.format.value]
        page = self.accesor.readPage(self.number)
        buf = np.array(page.getBuffer(), copy=False, dtype=dtype)
        return buf.__getitem__(key)

    def __setitem__(self, key: Sequence[Union[int, slice]], value: np.array):
        dtype = FORMAT2NPTYPE[self.format.value]
        buf = np.array(self.page.getWritableBuffer(), copy=False, dtype=dtype)
        return buf.__setitem__(key, value)

    def release(self) -> None:
        self.page.release()
        self.is_released = True

    @property
    def page(self):
        if self._page is None:
            try:
                self._page = self.accesor.createPage(self.number)
            except openvds.core.InvalidOperation as e:
                if e.args[0] == "Cannot create a page that already exists":
                    self._page = self.accesor.readPage(self.number)
                else:
                    raise e
        return self._page

    @property
    def minmax(self) -> Tuple[Sequence[int]]:
        _min, _max = self.page.getMinMax()
        _min = _min[:3][::-1]
        _max = _max[:3][::-1]
        return _min, _max

    @property
    def slices(self) -> Tuple[slice]:
        return tuple(slice(i, j) for i, j in zip(*self.minmax))


class VDSChunksGenerator:
    def __init__(
        self,
        chunks_count: int,
        accessor: openvds.core.VolumeDataPageAccessor,
        format: Formats
    ) -> None:
        self.chunks_count = chunks_count
        self.accessor = accessor
        self.format = format

    def __iter__(self):
        self.n = 0
        return self

    def __next__(self):
        if self.n < self.chunks_count:
            chunk = VDSChunk(
                number=self.n, accessor=self.accessor, format=self.format
            )
            self.n += 1
            return chunk
        else:
            raise StopIteration


class Axis:
    def __init__(
        self,
        samples: int,
        name: str,
        coordinate_min: float,
        coordinate_max: float,
        unit: str = "unitless",
    ) -> None:
        self.samples = samples
        self.name = name
        self.unit = unit
        self.coordinate_min = coordinate_min
        self.coordinate_max = coordinate_max

    def __repr__(self) -> str:
        return f"<Axis(name={self.name}, unit={self.unit}, range=({self.coordinate_min}, {self.coordinate_max}))>"


class Channel:
    def __init__(
            self,
            name: str,
            format: Formats,
            unit: str,
            value_range_min: float,
            value_range_max: float,
            components: Components,
            accessor: openvds.core.VolumeDataPageAccessor = None,
            chunks_count: int = None,
            vds_source=None,
            shape: Sequence[int] = None,
            dimensions_nd=Dimensions._012,
    ) -> None:
        self._vds_source = vds_source
        self.name = name
        self.format = format
        self.unit = unit
        self.accessor = accessor
        self.chunks_count = chunks_count
        self.shape = shape
        self.components = components
        self.value_range_min = value_range_min
        self.value_range_max = value_range_max
        self.dimensions_nd = dimensions_nd

    def __repr__(self) -> str:
        return f"<Channel(name={self.name}, unit={self.unit}, format={self.format.name})>"

    def chunks(self) -> VDSChunksGenerator:
        return VDSChunksGenerator(chunks_count=self.chunks_count, accessor=self.accessor, format=self.format)

    def get_chunk(self, number: int) -> VDSChunk:
        if number not in set(range(self.chunks_count)):
            raise VDSException(f"Chunk number is out of range of: 0 to {self.chunks_count-1}")
        return VDSChunk(
            number=number, accessor=self.accessor, format=self.format
        )

    def _read_data(
        self,
        vds_source: openvds.core.VDS,
        begin: Sequence[int],
        end: Sequence[int],
        lod: int = 0,
        replacementNoValue: float = 0.0,
        channel: int = 0
    ):
        begin = begin[::-1] + ([0]*len(begin))
        end = end[::-1] + ([1]*len(end))

        dims = (
            end[2] - begin[2],
            end[1] - begin[1],
            end[0] - begin[0],
        )
        accessManager = openvds.VolumeDataAccessManager(vds_source)
        req = accessManager.requestVolumeSubset(
            begin,  # start slice
            end,  # end slice
            format=self.format.value,
            lod=lod,
            replacementNoValue=replacementNoValue,
            channel=channel,
        )

        if req.data is None:
            err_code, err_msg = accessManager.getCurrentDownloadError()
            logger.exception(err_code)
            logger.exception(err_msg)
            raise RuntimeError(f"requestVolumeSubset failed! Message: {err_msg}, Error Code: {err_code}")

        return req.data.reshape(*dims)

    def _getitem_for_whole_dataset(self, key: Sequence[Union[int, slice]]) -> np.array:
        is_int = False
        if all([isinstance(i, int) for i in key]):
            begin = [i for i in key]
            end = [i+1 for i in key]
            return self._read_data(self._vds_source, begin, end).__getitem__(tuple(0 for i in key))
        elif all([isinstance(i, int) or isinstance(i, slice) for i in key]):
            begin = []
            end = []
            for i, k in enumerate(key):
                if isinstance(k, slice):
                    begin.append(k.start if k.start else 0)
                    end.append(k.stop if k.stop else self.shape[i])
                elif isinstance(k, int):
                    is_int = True
                    begin.append(k)
                    end.append(k+1)
        else:
            raise VDSException("Item key is not list of slices or int")

        if is_int:
            return self._read_data(self._vds_source, begin, end).__getitem__(
                tuple(
                    slice(None, None) if isinstance(i, slice) else 0 for i in key
                )
            )
        else:
            return self._read_data(self._vds_source, begin, end)

    def __getitem__(self, key: Sequence[Union[int, slice]]) -> np.array:
        return self._getitem_for_whole_dataset(key)

    def commit(self):
        self.accessor.commit()


class VDS:
    def __init__(
        self,
        path: str,
        connection_string: str = "",
        databrick_size: BrickSizes = BrickSizes._128,
        metadata_dict: MetadataContainer = {},
        channels_data: List[np.array] = None,
        channels: List[Channel] = None,
        axes: List[Axis] = None,
        lod: LOD = LOD._None,
        negative_margin: int = 0,
        positive_margin: int = 0,
        options: Options = Options._None,
        full_resolution_dimension: int = 0,
        brick_size_2d_multiplier: int = 4,
        init_value: InitValue = InitValue.omit_init,
        access_mode: AccessModes = AccessModes.ReadOnly
    ) -> None:
        super().__init__()

        if access_mode in {AccessModes.ReadOnly, AccessModes.ReadWrite, AccessModes.ReadWriteWithoutLODGeneration}:
            try:
                self._vds_source = openvds.open(path, connection_string)
                self.initialize(
                    path=path,
                    access_mode=access_mode,
                    connection_string=connection_string,
                )
            except RuntimeError as e:
                if str(e) == "Open error: File::open ":
                    raise VDSException(f"Could not open vds for path {path}")
        else:
            logger.debug("Creating new VDS source...")
            self._vds_source = self.create(
                path=path,
                connection_string=connection_string,
                databrick_size=databrick_size,
                channels=channels,
                axes=axes,
                metadata_dict=metadata_dict,
                channels_data=channels_data,
                access_mode=access_mode,
                lod=lod,
                positive_margin=positive_margin,
                negagitve_margin=negative_margin,
                options=options,
                full_resolution_dimension=full_resolution_dimension,
                brick_size_2d_multiplier=brick_size_2d_multiplier,
                init_value=init_value
            )

            self.initialize(
                path=path,
                access_mode=access_mode,
                connection_string=connection_string
            )

    def initialize(
        self,
        path: str,
        access_mode: AccessModes,
        connection_string: str = "",
    ):
        self.closed = False
        self.path = path
        self.connection_string = connection_string
        self._channels = {}
        self._axes = {}
        self.closed = False

        self._layout = openvds.getLayout(self._vds_source)
        self._dimensionality = self._layout.getDimensionality()

        vds_info = get_vds_info(path, connection_string)
        _info = vds_info['layoutInfo'] if 'layoutInfo' in vds_info else vds_info
        databrick_size = BrickSizes.get_from_info(_info)

        for i, j in enumerate(_info['axisDescriptors']):
            self._axes[j['name']] = Axis(
                samples=j['numSamples'],
                name=j['name'],
                unit=j['unit'],
                coordinate_max=j['coordinateMax'],
                coordinate_min=j['coordinateMin']
            )
        self.chunks_count = self.count_number_of_chunks(self.shape, databrick_size)
        for i, j in enumerate(_info['channelDescriptors']):

            if access_mode == AccessModes.Create:
                _access_mode = AccessModes.ReadWrite
            elif access_mode == AccessModes.CreateWithoutLODGeneration:
                _access_mode = AccessModes.ReadWriteWithoutLODGeneration
            else:
                _access_mode = access_mode

            self._channels[j['name']] = Channel(
                vds_source=self._vds_source,
                shape=self.shape,
                components=getattr(Components, j['components'].replace("Components", "")),
                name=j['name'],
                unit=j['unit'],
                format=getattr(
                    Formats, j['format'].replace("Format_", "")
                ),
                value_range_max=j['valueRange'][1],
                value_range_min=j['valueRange'][0],
                accessor=self._create_accessor(channel=i, access_mode=_access_mode),
                chunks_count=self.chunks_count
            )

    def channel(self, number: int) -> Channel:
        return self.channels[number]

    @property
    def channels(self) -> List[Channel]:
        return list(self._channels.values())

    @property
    def axes(self) -> List[Axis]:
        return list(self._axes.values())[::-1]

    def get_channel(self, name: str) -> Channel:
        return self._channels[name]

    def close(self, flush: bool = True):
        if not getattr(self, "closed", True):
            openvds.close(self._vds_source, flush)
            self.closed = True

    def __del__(self):
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.close()

    @property
    def metadata(self) -> MetadataContainer:
        return MetadataContainer.get_from_layout(self._layout)

    @property
    def negative_margin(self) -> int:
        return self._layout.getLayoutDescriptor().getNegativeMargin()

    @property
    def positive_margin(self) -> int:
        return self._layout.getLayoutDescriptor().getPositiveMargin()

    @property
    def full_resolution_dimension(self) -> int:
        return self._layout.getLayoutDescriptor().fullResolutionDimension

    @property
    def options(self):
        name = str(self._layout.getLayoutDescriptor().getOptions()).replace("Options.Options", "")
        return getattr(Options, name)

    @property
    def brick_size_2d_multiplier(self) -> int:
        return self._layout.getLayoutDescriptor().brickSizeMultiplier2D

    @property
    def databrick_size(self) -> BrickSizes:
        name = str(self._layout.getLayoutDescriptor().getBrickSize()).replace("BrickSize.BrickSize", "")
        return getattr(BrickSizes, name)

    @property
    def lod(self) -> LOD:
        name = str(self._layout.getLayoutDescriptor().getLODLevels()).replace("LODLevels.LODLevels", "")
        return getattr(LOD, name)

    def __str__(self) -> str:
        return f"<{self.__class__.__qualname__}(path={self.path})>"

    @property
    def axis_descriptors(self):
        return tuple((
            self._axes[k].name,
            self._axes[k].unit,
            self._axes[k].samples
        ) for k in self._axes)[::-1]

    @property
    def shape(self):
        return tuple(self._axes[k].samples for k in self._axes)[::-1]

    @staticmethod
    def create(
        path: str,
        connection_string: str,
        databrick_size: BrickSizes,
        access_mode: AccessModes,
        channels: List[Channel],
        axes: List[Axis],
        positive_margin: int,
        negagitve_margin: int,
        full_resolution_dimension: int,
        brick_size_2d_multiplier: int,
        lod: LOD,
        options: Options,
        init_value: InitValue,
        metadata_dict: MetadataContainer = None,
        channels_data: List[np.array] = None,
    ):
        return create_vds(
            path=path,
            connection_string=connection_string,
            metadata_dict=metadata_dict,
            channels=channels,
            axes=axes,
            databrick_size=databrick_size.value,
            access_mode=access_mode.value,
            lod=lod.value,
            channels_data=channels_data,
            negative_margin=negagitve_margin,
            positive_margin=positive_margin,
            options=options.value,
            brick_size_2d_multiplier=brick_size_2d_multiplier,
            full_resolution_dimension=full_resolution_dimension,
            init_value=init_value,
        )

    @property
    def accessor(self):
        if self._accessor is None:
            raise VDSException("Accessor was not initialized use create_accessor method.")
        return self._accessor

    def _create_accessor(
            self,
            access_mode: AccessModes = AccessModes.ReadWrite,
            lod: int = 0,
            channel: int = 0,
            maxPages: int = 8,
            chunkMetadataPageSize: int = 1024,
            dimensionsND=Dimensions._012,
    ):
        access_manager = openvds.getAccessManager(self._vds_source)
        accessor = access_manager.createVolumeDataPageAccessor(
            dimensionsND=dimensionsND.value,
            accessMode=access_mode.value,
            lod=lod,
            channel=channel,
            maxPages=maxPages,
            chunkMetadataPageSize=chunkMetadataPageSize,
        )
        self._accessor = accessor
        return self._accessor

    @staticmethod
    def count_number_of_chunks(shape: int, brick_size: BrickSizes):
        x = [round(i/(2**brick_size.value.value)) for i in shape]
        r = 1
        for i in x:
            if i != 0:
                r *= i
        return r

    def __getitem__(self, key: Sequence[Union[int, slice]]) -> np.array:
        return self.channel(0).__getitem__(key)


class VDSComposite:
    def __init__(self, subsets: Sequence[VDS] = None, slice_dim: int = None) -> None:
        if subsets is None:
            subsets = []
        if slice_dim is None:
            slice_dim = 0
        self.__subsets = subsets
        self.__shapes = []

        for s in self.__subsets:
            self.__shapes.append(s.shape)

        self.__slice_dim = slice_dim

        self.shape = None
        for i, s in enumerate(subsets):
            if i == 0:
                self.shape = list(s.shape)
            else:
                self.shape[slice_dim] += s.shape[slice_dim]

    def add_subset(self, subset: VDS):
        self.__subsets.append(subset)
        self.__shapes.append(subset.shape)

        self.shape = None
        for i, s in enumerate(self.__subsets):
            if i == 0:
                self.shape = list(s.shape)
            else:
                self.shape[self.__slice_dim] += s.shape[self.__slice_dim]

        self.shape = tuple(self.shape)

    def __getitem__(self, key: Sequence[Union[int, slice]]) -> np.array:
        is_int = False
        if all([isinstance(i, slice) or isinstance(i, int) for i in key]):
            if all([isinstance(i, int) for i in key]):
                cnt = 0
                _slice = key[self.__slice_dim]
                key = list(key)
                for i in range(len(self.__shapes)):
                    b = cnt + self.__shapes[i][self.__slice_dim]
                    if _slice >= cnt and _slice < b:
                        key[self.__slice_dim] = _slice - cnt
                        return self.__subsets[i].__getitem__(tuple(key))
                    cnt += self.__shapes[i][self.__slice_dim]
            else:
                _key = []
                for i, k in enumerate(key):
                    if isinstance(k, int):
                        is_int = True
                        _key.append(slice(k, k+1))
                    else:
                        _key.append(slice(k.start or 0, k.stop or 0, k.step))
                _key = tuple(_key)
        else:
            raise VDSException("Key elements must be instances of slice or int.")

        for i, k in enumerate(_key):
            if k.start is not None and (k.start > self.shape[i] + 1 or k.stop > self.shape[i] + 1):
                raise Exception(f"{_key} is out of range {self.shape}")

        _slice = _key[self.__slice_dim]
        keys = []
        a = 0
        for i in range(len(self.__shapes)):
            b = a + self.__shapes[i][0]
            if _slice.start < b and (_slice.start == _slice.stop):
                _slices = list(deepcopy(_key))
                _slices[self.__slice_dim] = slice(0, _slice.stop - a)
                keys.append(tuple(_slices))
                break
            elif _slice.start < b:
                _slices = list(deepcopy(_key))
                _slices[self.__slice_dim] = slice(
                    0,
                    self.__shapes[i][self.__slice_dim]
                )
                keys.append(tuple(_slices))
            else:
                keys.append(None)
            if _slice.stop < b:
                _slices = list(deepcopy(keys[-1]))
                _slices[self.__slice_dim] = slice(0, _slice.stop - a)
                keys[-1] = tuple(_slices)
                break
            a += self.__shapes[i][0]
        result_list = []
        for i, s in enumerate(keys):
            if s:
                result_list.append(self.__subsets[i].__getitem__(s))
        if is_int:
            return np.concatenate(tuple(result_list), axis=self.__slice_dim).__getitem__(
                tuple(
                    slice(None, None) if isinstance(i, slice) else 0 for i in key
                )
            )

        return np.concatenate(tuple(result_list), axis=self.__slice_dim)



from enum import Enum, auto

import openvds


class InitValue(Enum):
    NaN = auto()
    zero = auto()
    omit_init = auto()


class Formats(Enum):
    _1Bit = openvds.VolumeDataChannelDescriptor.Format.Format_1Bit
    U8 = openvds.VolumeDataChannelDescriptor.Format.Format_U8
    U16 = openvds.VolumeDataChannelDescriptor.Format.Format_U16
    R32 = openvds.VolumeDataChannelDescriptor.Format.Format_R32
    U32 = openvds.VolumeDataChannelDescriptor.Format.Format_U32
    U64 = openvds.VolumeDataChannelDescriptor.Format.Format_U64
    R64 = openvds.VolumeDataChannelDescriptor.Format.Format_R64


class Components(Enum):
    _1 = openvds.VolumeDataChannelDescriptor.Components.Components_1
    _2 = openvds.VolumeDataChannelDescriptor.Components.Components_2
    _4 = openvds.VolumeDataChannelDescriptor.Components.Components_4


class BrickSizes(Enum):
    _64 = openvds.core.VolumeDataLayoutDescriptor.BrickSize.BrickSize_64
    _128 = openvds.core.VolumeDataLayoutDescriptor.BrickSize.BrickSize_128
    _256 = openvds.core.VolumeDataLayoutDescriptor.BrickSize.BrickSize_256
    _512 = openvds.core.VolumeDataLayoutDescriptor.BrickSize.BrickSize_512
    _1024 = openvds.core.VolumeDataLayoutDescriptor.BrickSize.BrickSize_1024
    _2048 = openvds.core.VolumeDataLayoutDescriptor.BrickSize.BrickSize_2048
    _4096 = openvds.core.VolumeDataLayoutDescriptor.BrickSize.BrickSize_4096

    @classmethod
    def get_from_info(cls, vds_info: dict):
        return getattr(cls, "_"+vds_info["layoutDescriptor"]["brickSize"].rsplit("_")[-1])


class AccessModes(Enum):
    Create = openvds.IVolumeDataAccessManager.AccessMode.AccessMode_Create
    ReadOnly = openvds.IVolumeDataAccessManager.AccessMode.AccessMode_ReadOnly
    ReadWrite = openvds.IVolumeDataAccessManager.AccessMode.AccessMode_ReadWrite
    CreateWithoutLODGeneration = openvds.IVolumeDataAccessManager.AccessMode.AccessMode_CreateWithoutLODGeneration
    ReadWriteWithoutLODGeneration = openvds.IVolumeDataAccessManager.AccessMode.AccessMode_ReadWriteWithoutLODGeneration


class Dimensions(Enum):
    _012 = openvds.DimensionsND.Dimensions_012


class Options(Enum):
    _None = openvds.VolumeDataLayoutDescriptor.Options.Options_None
    _2DLODs = openvds.VolumeDataLayoutDescriptor.Options.Options_Create2DLODs


class LOD(Enum):
    _None = openvds.VolumeDataLayoutDescriptor.LODLevels.LODLevels_None
    _1 = openvds.VolumeDataLayoutDescriptor.LODLevels.LODLevels_1
    _2 = openvds.VolumeDataLayoutDescriptor.LODLevels.LODLevels_2
    _3 = openvds.VolumeDataLayoutDescriptor.LODLevels.LODLevels_3
    _4 = openvds.VolumeDataLayoutDescriptor.LODLevels.LODLevels_4
    _5 = openvds.VolumeDataLayoutDescriptor.LODLevels.LODLevels_5
    _6 = openvds.VolumeDataLayoutDescriptor.LODLevels.LODLevels_6
    _7 = openvds.VolumeDataLayoutDescriptor.LODLevels.LODLevels_7
    _8 = openvds.VolumeDataLayoutDescriptor.LODLevels.LODLevels_8
    _9 = openvds.VolumeDataLayoutDescriptor.LODLevels.LODLevels_9
    _10 = openvds.VolumeDataLayoutDescriptor.LODLevels.LODLevels_10
    _11 = openvds.VolumeDataLayoutDescriptor.LODLevels.LODLevels_11
    _12 = openvds.VolumeDataLayoutDescriptor.LODLevels.LODLevels_12


class MetadataTypes(Enum):
    IntVector2 = "MetadataType.IntVector2"
    Float = "MetadataType.Float"
    DoubleVector2 = "MetadataType.DoubleVector2"
    String = "MetadataType.String"
    Int = "MetadataType.Int"
    Double = "MetadataType.Double"
    BLOB = "MetadataType.BLOB"



from .copy import copy_vds  # NOQA
from .enums import (LOD, AccessModes, BrickSizes, Components, Dimensions, Formats, InitValue, MetadataTypes,  # NOQA
                    Options)
from .utils import METADATATYPE_TO_OVDS_GET_FUNCTION, METADATATYPE_TO_OVDS_SET_FUNCTION  # NOQA
from .writing import create_vds, write_pages  # NOQA



from __future__ import annotations

from subprocess import PIPE, Popen

from ovds_utils.logging import get_logger

from .utils import get_ovds_bin

logger = get_logger(__name__)


def copy_vds(
    source: str,
    target: str,
    source_connection: str = "",
    target_connection: str = "",
    additional_params={"tolerance": 1, "compression-method": None}
):
    cmd = [
        "OPENVDS_AWSCURL=1",
        get_ovds_bin("VDSCopy"),
        source,
        target,
        f"-s '{source_connection}'",
        f"-d '{target_connection}'",
    ]
    for k in additional_params:
        cmd.append(f"--{k} {additional_params[k]}")

    logger.info("Executing VDSCopy")
    process = Popen(
        " ".join(cmd), stdout=PIPE, stderr=PIPE, shell=True, universal_newlines=True
    )
    while process.poll() is None:
        info = process.stdout.readline()
        logger.info("VDSCopy:" + info)



from __future__ import annotations

from logging import getLogger
from typing import Any, AnyStr, Dict, List

import numpy as np
import openvds

from .enums import AccessModes, InitValue
from .utils import copy_ovds_metadata

logger = getLogger(__name__)

FORMAT2NPTYPE = {
    openvds.VolumeDataChannelDescriptor.Format.Format_R64: np.float64,
    openvds.VolumeDataChannelDescriptor.Format.Format_R32: np.float32,
    openvds.VolumeDataChannelDescriptor.Format.Format_U8: np.uint8,
    openvds.VolumeDataChannelDescriptor.Format.Format_U16: np.uint16,
    openvds.VolumeDataChannelDescriptor.Format.Format_U32: np.uint32,
    openvds.VolumeDataChannelDescriptor.Format.Format_U64: np.uint64
}


def write_pages(
    accessor: openvds.core.VolumeDataPageAccessor, data: np.array,
    format: openvds.VolumeDataChannelDescriptor.Format
):
    dtype = FORMAT2NPTYPE[format]
    for c in range(accessor.getChunkCount()):
        page = accessor.createPage(c)
        buf = np.array(page.getWritableBuffer(), copy=False, dtype=dtype)
        (min, max) = page.getMinMax()
        buf[:, :, :] = data[
            min[2]: max[2],
            min[1]: max[1],
            min[0]: max[0],
        ]
        page.release()
    accessor.commit()


def write_nan_pages(
    accessor: openvds.core.VolumeDataPageAccessor,
    format: openvds.VolumeDataChannelDescriptor.Format
):
    dtype = FORMAT2NPTYPE[format]
    for c in range(accessor.getChunkCount()):
        page = accessor.createPage(c)
        buf = np.array(page.getWritableBuffer(), copy=False, dtype=dtype)
        buf[:, :, :] = np.full(buf.shape, np.nan, dtype=dtype)
        page.release()
    accessor.commit()


def write_zero_pages(
    accessor: openvds.core.VolumeDataPageAccessor,
    format: openvds.VolumeDataChannelDescriptor.Format
):
    dtype = FORMAT2NPTYPE[format]
    for c in range(accessor.getChunkCount()):
        page = accessor.createPage(c)
        buf = np.array(page.getWritableBuffer(), copy=False, dtype=dtype)
        buf[:, :, :] = np.zeros(buf.shape, dtype=dtype)
        page.release()
    accessor.commit()


INITVALUE = {
    InitValue.NaN: write_nan_pages,
    InitValue.zero: write_zero_pages
}


def create_vds_attributes(
    databrick_size: openvds.core.VolumeDataLayoutDescriptor.BrickSize,
    metadata_dict: Dict[AnyStr, Any],
    lod_levels: openvds.VolumeDataLayoutDescriptor.LODLevels,
    options: openvds.VolumeDataLayoutDescriptor.Options,
    negative_margin: int,
    positive_margin: int,
    brick_size_2d_multiplier: int,
    full_resolution_dimension: int,
    channles,
    axes
):
    layout_descriptor = openvds.VolumeDataLayoutDescriptor(
        brickSize=databrick_size,
        lodLevels=lod_levels,
        brickSize2DMultiplier=brick_size_2d_multiplier,
        options=options,
        negativeMargin=negative_margin,
        positiveMargin=positive_margin,
        fullResolutionDimension=full_resolution_dimension,
    )
    metadata_container = openvds.MetadataContainer()
    copy_ovds_metadata(metadata_dict, metadata_container)
    axis_descriptors = []
    for a in axes[::-1]:
        axis_descriptors.append(
            openvds.VolumeDataAxisDescriptor(
                a.samples,
                a.name,
                a.unit,
                a.coordinate_min,
                a.coordinate_max
            )
        )

    channel_descriptors = []
    for c in channles:
        channel_descriptors.append(
            openvds.VolumeDataChannelDescriptor(
                format=c.format.value,
                components=c.components.value,
                name=c.name,
                unit=c.unit,
                valueRangeMin=c.value_range_min,
                valueRangeMax=c.value_range_max
            )
        )
    return layout_descriptor, axis_descriptors, channel_descriptors, metadata_container


def create_vds(
    path: AnyStr,
    connection_string: AnyStr,
    metadata_dict: Dict[AnyStr, Any],
    channels: List,
    axes: List,
    negative_margin: int,
    positive_margin: int,
    databrick_size: openvds.VolumeDataLayoutDescriptor.BrickSize,
    access_mode: openvds.IVolumeDataAccessManager.AccessMode,
    lod: openvds.VolumeDataLayoutDescriptor.LODLevels,
    options: int,
    brick_size_2d_multiplier: int,
    full_resolution_dimension: int,
    default_max_pages: int = 8,
    channels_data=None,
    init_value: InitValue = InitValue.zero,
):
    (
        layout_descriptor,
        axis_descriptors,
        channel_descriptors,
        metadata_container
    ) = create_vds_attributes(
        databrick_size=databrick_size,
        metadata_dict=metadata_dict,
        channles=channels,
        axes=axes,
        lod_levels=lod,
        negative_margin=negative_margin,
        positive_margin=positive_margin,
        options=options,
        brick_size_2d_multiplier=brick_size_2d_multiplier,
        full_resolution_dimension=full_resolution_dimension,
    )

    vds = openvds.create(
        url=path,
        connectionString=connection_string,
        layoutDescriptor=layout_descriptor,
        axisDescriptors=axis_descriptors,
        channelDescriptors=channel_descriptors,
        metadata=metadata_container,
    )
    access_manager = openvds.getAccessManager(vds)

    if channels_data:
        for i, data in enumerate(channels_data):
            channel = channels[i]
            accessor = access_manager.createVolumeDataPageAccessor(
                dimensionsND=channel.dimensions_nd.value,
                accessMode=AccessModes.Create.value,
                lod=lod,
                channel=i,
                maxPages=default_max_pages,
            )
            write_pages(accessor, data, channel.format.value)

    if init_value != InitValue.omit_init:
        for i in range(len(channels)):
            channel = channels[i]
            accessor = access_manager.createVolumeDataPageAccessor(
                dimensionsND=channel.dimensions_nd.value,
                accessMode=access_mode,
                lod=lod,
                channel=i,
                maxPages=default_max_pages,
            )
            INITVALUE[init_value](accessor, channel.format.value)

    return vds



from __future__ import annotations

import logging
import os
import sys

LOG_LEVEL = os.environ.get("LOGLVL", "NOTSET")
FORMAT = "%(asctime)s::%(name)s::%(levelname)s::%(message)s"


def get_logger(name) -> logging.Logger:
    logger = logging.getLogger(name)
    console_handler = logging.StreamHandler(sys.stdout)
    logger.setLevel(LOG_LEVEL)
    console_handler.setLevel(LOG_LEVEL)
    formatter = logging.Formatter(FORMAT)
    console_handler.setFormatter(formatter)

    logger.addHandler(console_handler)
    return logger



from __future__ import annotations

from typing import Any, AnyStr, Dict, Union

import openvds

from ovds_utils.exceptions import VDSMetadataException
from ovds_utils.ovds import METADATATYPE_TO_OVDS_GET_FUNCTION, METADATATYPE_TO_OVDS_SET_FUNCTION, MetadataTypes


class MetadataValue:
    def __init__(self, value: Any, category: AnyStr, type: Union[AnyStr, MetadataTypes]) -> None:
        self.value = value
        self.category = category
        if isinstance(type, MetadataTypes):
            self._type = type
        elif str(type) not in METADATATYPE_TO_OVDS_GET_FUNCTION:
            raise VDSMetadataException(
                f"The type {type} was not recognized among: {', '.join(METADATATYPE_TO_OVDS_GET_FUNCTION.keys())}")
        else:
            self._type = getattr(MetadataTypes, str(type).replace("MetadataType.", ""))

    def __repr__(self) -> str:
        return f"<{self.__class__.__qualname__}(value={self.value}, category={self.category}, type={self.type})>"

    @property
    def type(self):
        return self._type.value


class MetadataContainer(dict):
    def __init__(self, **kwargs: Dict[AnyStr, MetadataValue]) -> None:
        super().__init__()
        self.update(kwargs)

    def __repr__(self) -> str:
        return f"<{self.__class__.__qualname__}({', '.join(self.keys())})>"

    def get_container(self):
        container = openvds.MetadataContainer()
        for k, v in self.items():
            set_method = METADATATYPE_TO_OVDS_SET_FUNCTION[v.type]
            set_method(container, v.category, k, v.value)
        return container

    @staticmethod
    def get_from_layout(layout: openvds.core.VolumeDataLayout) -> openvds.core.MetadataContainer:
        metadata = {}
        for i in layout.getMetadataKeys():
            method = METADATATYPE_TO_OVDS_GET_FUNCTION[str(i.type)]
            value = method(layout, i.category, i.name)
            metadata[i.name] = MetadataValue(value, i.category, i.type)

        return MetadataContainer(**metadata)
